/* Copyright (c) 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Management of game modules. */

#include "conq.h"

/* List of all known modules. Their descriptions can co-exist in memory,
   even if their contents cannot. */

Module *modulelist = NULL;

/* The main module defining the game in effect. */

Module *mainmodule = NULL;

/* Empty out the list of modules. */

clear_game_modules()
{
    modulelist = mainmodule = NULL;
}

/* Create a brand-new game module. */

Module *
create_game_module(name)
char *name;
{
    Module *module = (Module *) xmalloc(sizeof(Module));

    /* Nullify/zeroify most module slots. */
    bzero(module, sizeof(Module));
    /* Fill in some others. */
    if (name == NULL) name = "";  /* a hack */
    module->name = name;
    module->instructions = lispnil;
    module->variants = lispnil;
    module->notes = lispnil;
    module->designnotes = lispnil;
    module->startlineno = 1;
    module->endlineno = 1;
    module->maybereshape = FALSE;
    /* Add to front of module list. */
    module->next = modulelist;
    modulelist = module;
    return module;
}

/* Produce a module of the given name, either by finding it or creating it. */

Module *
get_game_module(name)
char *name;
{
    Module *module;

    if (name != NULL) {
	for_all_modules(module) {
	    if (module->name && strcmp(name, module->name) == 0) return module;
	}
    }
    return create_game_module(name);
}

/* Make a game module for the given name and bolt it into the include
   list of another module, or else into the global list of modules. */

Module *
add_game_module(name, includer)
char *name;
Module *includer;
{
    Module *module = get_game_module(name);

    if (includer) {
	/* Add to the end of the list of include files. */
	if (includer->include == NULL) {
	    includer->include = includer->lastinclude = module;
	} else {
	    includer->lastinclude->nextinclude = module;
	    includer->lastinclude = module;
	}
    } else {
	/* an error? */
    }
    return module;
}

/* Display game module info to a side. */

describe_game_modules(arg, key, buf)
int arg;
char *key, *buf;
{
    if (mainmodule != NULL) {
	sprintf(buf, "Modules:\n\n");
	describe_game_module_aux(buf, mainmodule, 0);
	/* Now do the lengthy module notes (with no indentation). */
	describe_module_notes(buf, mainmodule);
    } else {
	sprintf(buf, "(No game module information is available.)");
    }
}

/* Recurse down through included modules to display docs on each.
   Indents each file by inclusion level.  Note that modules cannot
   be loaded more than once, so each will be described only once here. */
   
static char indentbuf[100];

describe_game_module_aux(buf, origmodule, indent)
char *buf;
Module *origmodule;
int indent;
{
    int i;
    Module *module;

    indentbuf[0] = '\0';
    for (i = 0; i < indent; ++i) {
	strcat(indentbuf, "  ");
    }
    /* A brief list of all the modules, indented to show nesting. */
    for_all_includes(origmodule, module) {
	sprintf(buf+strlen(buf), "%s\"%s\"\n",
		indentbuf, module->name);
	sprintf(buf+strlen(buf), "%s          %s\n",
		indentbuf,
		(module->blurb ? module->blurb : "(no description)"));
	if (module->notes != lispnil) {
	    sprintf(buf+strlen(buf),
		    "%s          (See notes below)\n", indentbuf);
	}
	if (module->include) {
	    sprintf(buf+strlen(buf), "%s \n", indentbuf);
	    describe_game_module_aux(buf, module->include, indent + 1);
	}
	/* should mention the various flags. */
    }
}

/* Dump the module designer's notes into the given buffer. */

describe_module_notes(buf, origmodule)
char *buf;
Module *origmodule;
{
    Obj *notes;
    Module *module;

    if (origmodule->notes != lispnil) {
	sprintf(buf+strlen(buf), "\nNotes to the Module \"%s\":\n",
		origmodule->name);
	for (notes = origmodule->notes; notes != lispnil; notes = cdr(notes)) {
	    strcat(buf, c_string(car(notes)));
	    strcat(buf, " ");
	}
    }
    for_all_includes(origmodule, module) {
	describe_module_notes(buf, module);
    }
}

/* Sometimes we find ourselves lacking a period to provide meaning and
   context for interpretation; this routine loads the standard game in
   immediately, but only makes it the main module if none defined. */

load_standard_game()
{
    extern char *standardgamename;
    Module *module = get_game_module(standardgamename);

    if (mainmodule == NULL) mainmodule = module;
    load_game_module(module);
}

/* Attempt to read just the first form in a module and use it as a
   description of the module.  Return true if this worked, false
   otherwise. */

load_game_description(module)
Module *module;
{
    Obj *form;
    Obj *thecar, *rest;
    char *name, *str;

    if (open_module(module, FALSE)) {
	if ((form = read_form(module->fp,
			      &(module->startlineno),
			      &(module->endlineno)))
	    != lispeof) {
	    if (consp(form) && symbolp(thecar = car(form))) {
		name = c_string(thecar);
		if (strcmp(name, "game-module") == 0) {
		    interp_game_module(form, module);
		    close_module(module);
		    /* Note that module is still not considered "loaded". */
		    return TRUE;
		}
	    }
	}
    }
    return FALSE;
}

/* Game files can live in library directories or somewhere else.  This
   function tries to find a file, open it, and load the contents. */

load_game_module(module)
Module *module;
{
    char ch;

    if (open_module(module, TRUE)) {
	if (module->fp) {
	    /* Peek at the first character - was 'X' in old format files. */
	    ch = getc(module->fp);
	    ungetc(ch, module->fp);
	    if (ch == 'X') {
		init_error("\"%s\" is probably an obsolete Xconq file; in any case, it cannot be used.",
			   module->name);
	    } else {
		/* New format, read it all. */
		read_forms(module);
	    }
	} else {
	    /* (should be able to read from contents string) */
	}
	/* We're done, can close. */
	close_module(module);
	module->loaded = TRUE; /* maybe only if no errors? */
	/* Make all the cross-references right. */
	patch_object_references();
    }
}

/* Given a module, attempt to open it. */

open_module(module, dowarn)
Module *module;
int dowarn;
{
    FILE *fp = NULL;

    /* Don't open more than once. */
    if (module->open && dowarn) {
	init_warning("\"%s\" is already open, ignoring this attempt to open.",
		     module->name);
	return FALSE;
    }
    /* Don't open if already loaded. */
    if (module->loaded && dowarn) {
	init_warning("\"%s\" is already loaded, ignoring this attempt to open.",
		     module->name);
	return FALSE;
    }
    if (module->contents) {
	/* Uninterpreted contents already available, init the ptr. */
	module->sp = module->contents;
	Dprintf("Reading module \"%s\" from string ...\n", module->name);
    } else if ((fp = open_library_file(module)) != NULL) {
	/* Found the module in a library directory. */
	Dprintf("Reading module \"%s\" from library file \"%s\" ...\n",
		module->name, module->filename);
	module->fp = fp;
    } else if ((fp = open_explicit_file(module)) != NULL) {
	Dprintf("Reading module \"%s\" from file \"%s\" ...\n",
		module->name, module->filename);
	module->fp = fp;
    } else {
	if (dowarn) {
	    if (module->name) {
	    	init_warning("Can't find module \"%s\" anywhere!", module->name);
	    } else {
	    	init_warning("Can't find unnamed module anywhere!");
	    }
	}
	return FALSE;
    }
    /* It worked, mark this module as open. */
    module->open = TRUE;
    return TRUE;
}

/* Read info about a side's preferences and setup. */

/* This assumes one form only, probably too restrictive. */
/* should read all the forms, use the relevant ones. */
/* (how does this interact with other defaults?) */
/* (should be delayed until player can confirm it...) */

/* (update to work like other module stuff? then can use resources etc) */
/* (fix so that correct name can be found reliably) */

load_side_config(side)
Side *side;
{
#if 0
    FILE *fp;
    Obj *config;
    Module *module;

    /* (should incorp config name somehow, also be sys-dependent) */
    module = create_game_module(side->player->name);

    if ((module->fp = fopen(module->filename, "r")) != NULL) {
	if ((config = read_form(module->fp,
				&(module->startlineno),
				&(module->endlineno)))
	    != lispeof) {
	    /* interpret the config */
	    Dprintf("Interpreting %s config form", side_desig(side));
	    Dprintlisp(config);
	    Dprintf("\n");
	    fill_in_side(side, config, TRUE);
	} else {
	    /* no config form in the file */
	}
    } else {
	init_warning("\"%s\" could not be opened!", module->name);
	/* Not a disaster, keep going */
    }
#endif
}

/* Read an entire file, attempting to pick up objects in it. */

read_forms(module)
Module *module;
{
    Obj *form;

    Dprintf("Trying to read a new format file \"%s\"...\n", module->name);
    while ((form = read_form(module->fp,
			     &(module->startlineno),
			     &(module->endlineno)))
	   != lispeof) {
	interp_form(module, form);
    }
    Dprintf("... Done reading \"%s\".\n", module->name);
}

init_module_reshape(module)
Module *module;
{
    /* Seed all the reshaping parameters with reasonable values. */
    module->maybereshape = TRUE;
    module->subareawidth = area.width;
    module->subareaheight = area.height;
    module->subareax = module->subareay = 0;
    module->finalsubareawidth = area.width;
    module->finalsubareaheight = area.height;
    module->finalsubareax = module->finalsubareay = 0;
    module->finalwidth = area.width;  module->finalheight = area.height;
    module->finalcircumference = world.circumference;
    module->filltype = 0;
}

/* This is true if any actual reshaping is required. */

reshape_the_output(module)
Module *module;
{
    return (module->maybereshape
	    && (module->subareawidth != area.width
		|| module->subareaheight != area.height
		|| module->subareax != 0
		|| module->subareay != 0
		|| module->finalsubareawidth != area.width
		|| module->finalsubareaheight != area.height
		|| module->finalsubareax != 0
		|| module->finalsubareay != 0
		|| module->finalwidth != area.width
		|| module->finalheight != area.height
		|| module->finalcircumference != world.circumference));
}

/* Check if the proposed reshape will actually work. */

valid_reshape(module)
Module *module;
{
	/* (should check hexagon shaping) */
	if (module->subareawidth > area.width || module->subareaheight > area.height) return FALSE;
	/* (should check other offsets) */
	if (module->finalwidth < 3 || module->finalheight < 3) return FALSE;
	return TRUE;
}

/* Close the module. */

close_module(module)
Module *module;
{
    if (module->sp) {
	module->sp = NULL;
    }
    if (module->fp) {
	fclose(module->fp);
	module->fp = NULL;
    }
    module->open = FALSE;
}

/* Dump a description of the module into a static buffer. */

char moduledesigbuf[BUFSIZE];

char *
module_desig(module)
Module *module;
{
    sprintf(moduledesigbuf, "module %s (%s)",
	    module->name, (module->title ? module->title : "no title"));
    return moduledesigbuf;
}
