/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This is the main simulation-running code. */

#include "conq.h"

/* The number of the current turn. */

int curturn;

/* Sequencing number within the turn. */

int cursequence;

Unit *curgroup = NULL;

Unit *lastunit = NULL;

Obj *randomeventlist;
Obj *lastrandomevent;

int anyrandomevents = FALSE;

int run_accidents();
int run_attrition();
int run_revolts();

struct randomeventtype {
    char *name;
    int (*fn)();
    int dflt;
    int run;
} randomeventmethods[] = {
    { "accidents-in-terrain", run_accidents, FALSE, FALSE },
    { "attrition-in-terrain", run_attrition, FALSE, FALSE },
    { "random-revolts", run_revolts, FALSE, FALSE },
    { NULL, NULL, 0 }
};

int maintimeout = -1;

int paused = FALSE;

/* State variables. */
/* (I don't think all of these are strictly necessary) */

/* This is true when we are actually actively running the game. */

int ingame = FALSE;

int beforestart = TRUE;
int startturn = FALSE;
int movestuff = FALSE;
int finishturn = FALSE;
int endofgame = FALSE;

/* How often to do saves while playing. */

int checkpointinterval = 0;

/* This is set FALSE whenever the game state changes, and TRUE whenever the game
   has been saved. */

int gamestatesafe = TRUE;

int daynight = FALSE;

int sunx, suny;

/* This function does a (small usually) amount of simulation, then returns.
   It can be run multiple times at any time, will not go "too far".
   It returns the max time in seconds that the interface should do
   things before running this again. */

run_game(maxactions)
int maxactions;
{
    int timeout = -1, nummoved;
    Side *side;

    if (!ingame) {
	/* Still waiting to get a game set up. */
    } else if (beforestart) {
	/* If we haven't started yet, see if it's time. */
	test_for_game_start();
    } else if (endofgame) {
	/* Nothing to do except wait for users to do exit commands. */
    } else if (paused) {
	/* Don't do anything if we're paused. */
    } else {
	/* shouldn't do if game being restored... */
	/* actually, if move allowances were restored. */
	if (startturn) {
	    run_new_turn();
	    startturn = FALSE;
	}
	/* Non-participating side are automatically "finished". */
	for_all_sides(side) {
	    if (!side->finishedturn
		&& !side_has_ai(side)
		&& !side_has_display(side)) {
		finish_turn(side);
	    }
	}
	/* If all sides are done acting, end the turn.  This won't be true
	   right at the start of a turn. */
	if (all_sides_finished() /* or timed out or whatever */) {
	    curgroup = lastunit = NULL;
	    finish_movement();
	    run_base_consumption();
	    flush_dead_units();
	    check_turn_end_scores();
	    test_agreements();
	    if (Debug) report_malloc();
	    startturn = TRUE;
	} else {
	    /* Move some units around. */
	    nummoved = move_some_units(maxactions);
	    for_all_sides(side) {
		if (!side->finishedturn) {
		    /* Display and/or AI might have vanished? */
		    if (!side_has_ai(side) && !side_has_display(side)) {
			finish_turn(side);
		    }
		    /* See if any sides auto-finish. */
		    if (!units_still_acting(side)
			&& side->autofinish
			&& !side->designer) {
		    	Dprintf("%s auto-finishes.\n", side_desig(side));
			finish_turn(side);
		    }
		}
	    }
	}
	/* Decide on the timeout for the next call of this routine. */
	timeout = 0;
	/* Sides that have no-activity-timeouts will adjust the
	   timeout to occur at the end of the latest timeout; sides
	   that have explicitly declared themselves done won't need
	   any additional waits. */
	for_all_sides(side) {
	    if (!side->finishedturn) {
		timeout = max(timeout,
			      (side->realtimeout
			       - (time(0) - side->lasttime)));
	    }
	}
	test_for_game_end();
	return 1000 * timeout;
    }
    return timeout;
}

/* See if game is ready to get underway for real.  Note that everything
   "works" before the game has started, but time doesn't move. */

test_for_game_start()
{
    int onedisplay = FALSE;
    Side *side;

    /* We have to have at least one unit on a side that is being displayed. */
    for_all_sides(side) {
	if (side_has_display(side)) {
	    onedisplay = TRUE;
	}
	if (side_has_units(side) && side_has_display(side)) {
	    /* Now we're really ready to roll. */
	    beforestart = FALSE;
	    startturn = TRUE;
	    record_event(H_GAME_STARTED);
	    /* Record the game as starting NOW, in real time. */
	    if (g_rt_for_game_max() > 0) {
		record_real_start_time();
	    }
	    return;
	}
    }
    if (!onedisplay) {
	init_warning("No sides have a display!");
    }
}

/* This routine looks to see if the game is completely over. */

test_for_game_end()
{
    Side *side;

    for_all_sides(side) {
	if (side_has_display(side) && side->ingame) return;
    }
    endofgame = TRUE;
}

#if 0
	/* Add this time interval to all sides's time consumption. */
	for_all_sides(side) {
	    int delta_time = (time(0) - side->lasttime);

	    if (side->more_units) {
		side->timeleft -= delta_time;
		side->timetaken += delta_time;
		update_clock_display(side, TRUE);
	    }
	}
#endif

/* This is true when all participating sides have finished their turn. */

all_sides_finished()
{
    Side *side;

    for_all_sides(side) {
	if (side->ingame && !side->finishedturn) {
	    return FALSE;
	}
    }
    return TRUE;
}

/* This returns true if the given side is still wanting to do stuff. */

units_still_acting(side)
Side *side;
{
    Unit *unit;

    if (side->ingame) {
	for_all_side_units(side, unit) {
	    if (unit_still_acting(unit, FALSE)) {
		return TRUE;
	    }
	}
    }
    return FALSE;
}

/* Do everything that would happen before movement in a turn. */

run_new_turn()
{
    Side *side;

    if (!midturnrestore) {
	/* Increment the turn number. */
	++curturn;
	cursequence = 0;
	Dprintf("##### TURN %d (%s) #####\n",
		curturn, absolute_date_string(curturn));
	for_all_sides(side) {
	    side->finishedturn = FALSE;
	    update_turn_display(side, TRUE);
	    if (realtime_game()) {
		update_clock_display(side, TRUE);
	    }
	}
	run_sun();
	run_environment();
	run_economy();
	run_self_builds();
	run_reinforcements();
	run_random_events();
	sort_units();
	init_action_order();
	compute_moves();
	run_spies();
	cache_init_tech_levels();
/*	reset_coverage();  */
	gamestatesafe = FALSE;
    }
    if ((checkpointinterval > 0) && (curturn % checkpointinterval == 0)) {
	write_entire_game_state(checkpoint_name());
    }
    init_movement();
}

init_movement()
{
    Side *side;

    for_all_sides(side) {
	if (side->ingame) {
	    /* No units are waiting for orders initially. */
	    side->numwaiting = 0;
	}
    }
    for_all_sides(side) {
	side->turnstarttime = time(0);
	/* Didn't really do input, but useful to pretend so. */
	side->lasttime = time(0);
	if (side_has_ai(side)) {
	    ai_init_turn(side);
	}
	side->busy = FALSE;
	/* Inform sides with displays that their units are ready to act. */
	if (side_has_display(side)) {
	    update_action_display(side, TRUE);
	}
    }
}

/* Remember each side's tech levels before it does any research actions
   during the turn.  This is used to keep tech level from going up too
   fast if the player has lots of units doing research. */

cache_init_tech_levels()
{
    int u;
    Side *side;

    if (using_tech_levels() /* and any limits on research gains? */) {
	for_all_sides(side) {
	    for_all_unit_types(u) {
		side->inittech[u] = side->tech[u];
	    }
	}
    }
}

/* Clean out all details about what order units will move in. */

init_action_order()
{
    Unit *unit;
    Side *side;

    for_all_sides_plus_indep(side) {
	side->actorder = NULL;
    }
    for_all_units(side, unit) {
	if (unit->act != NULL) {
	    unit->nextactor = unit->prevactor = NULL;
	    unit->nextgroup = NULL;
	}
	/* Units start a turn *not* in reserve. */
	if (unit->plan != NULL) {
	    unit->plan->reserve = FALSE;
	}
    }
}

/* Compute moves and actions for all the units at once, put everybody that
   can do anything into a list. */

compute_moves()
{
    Unit *unit;
    Side *side;

    curgroup = lastunit = NULL;
    if (midturnrestore) { /* not necessary because already made? */
	midturnrestore = FALSE;
	/* but need to arrange numacting etc? */
    } else { /* for all sided units (why?). */
	for_all_sides(side) {
	    side->numacting = 0;
	    side->numfinished = 0;
	    side->actorder = NULL;
	    for_all_side_units(side, unit) {
		if (unit->act) {
		    compute_acp(unit);
		    compute_mp(unit);
		}
		/* Add unit to the list of actors. */
		if (unit->act != NULL && unit->act->acp > 0) {
		    if (side->actorder == NULL) side->actorder = unit;
		    if (curgroup == NULL) {
			curgroup = lastunit = unit;
		    } else {
			lastunit->nextactor = unit;
			unit->nextactor = NULL;
			lastunit = unit;
		    }
		    ++(side->numacting);
		}
	    }
	}
    }
}

/* Compute the action points available to the unit this turn. */

compute_acp(unit)
Unit *unit;
{
    int u = unit->type, maxacp;
    Unit *occ;

    /* Units still under construction or off-area can't do anything. */
    if (!completed(unit) || !inside_area(unit->x, unit->y)) {
	unit->act->initacp = unit->act->acp = unit->act->actualactions = 0;
	return;
    }
    /* First compute how many action points are available. */
    unit->act->initacp = unit->act->acp + u_acp(u);
    /* (should add effect of occupants if defined) */
    maxacp = (u_acp_max(u) < 0 ? u_acp(u) : u_acp_max(u));
    unit->act->initacp = min(unit->act->initacp, maxacp);
    unit->act->initacp = max(unit->act->initacp, u_acp_min(u));
    unit->act->acp = unit->act->initacp;
    unit->act->actualactions = 0;
}

/* Compute the movement points available to the unit this turn. */

compute_mp(unit)
Unit *unit;
{
    int u = unit->type, m, moves, maxacp;
    Unit *occ;

    /* (do incomplete units still need mp in order to be moved around?) */
    if (!completed(unit) || !inside_area(unit->x, unit->y)) {
	unit->act->initmp = unit->act->mp = unit->act->actualmoves = 0;
	return;
    }
    /* Start with the maximum speed. */
    moves = u_speed(u);
    /* Compute any effect of damage on the unit's speed. */
    if (moves > 0) {
	if (unit->hp < u_hp_at_max_speed(u)) {
	    if (unit->hp <= u_hp_at_min_speed(u)) {
		moves = u_speed_min(u);
	    } else if (u_hp_at_min_speed(u) == u_hp_at_max_speed(u)) {
		/* ? */
		moves = -1;
	    } else {
		/* Interpolate to get speed, rounding down. */
		moves =
		  u_speed_min(u) + (u_speed(u) - u_speed_min(u)) /
		    (u_hp_at_max_speed(u) - u_hp_at_min_speed(u));
	    }
	}
	/* Compute any effect of occupants on a transport's speed. */
	for_all_occupants(unit, occ) {
	    if (uu_mobility(u, occ->type) != 100) {
		moves = (moves * uu_mobility(u, occ->type)) / 100;
		break;  /* or make conditional somehow? */
	    }
	}
	/* The previous slowdowns should never reduce below 1. */
	moves = max(1, moves);
	/* Missing movement materials can immobilize us however. */
	/* (need to warn players when this happens) */
	for_all_material_types(m) {
	    if (um_tomove(u, m) > 0 && unit->supply[m] <= 0) {
		moves = 0;
		break;
	    }
	}
    }
    /* Set the available moves.  Note that since these are not accumulated
       like action points are, you can't save up moves from turn to turn. */
    unit->act->initmp = unit->act->mp = moves;
    unit->act->actualmoves = 0;
}

/* Do some number of actions. */

move_some_units(lim)
int lim;
{
    int num = 0;
    Unit *unit;
    Side *loopside;

    for_all_units(loopside, unit) {
	if (unit_still_acting(unit, TRUE)) {
	    num += move_one_unit_multiple(unit, lim - num);
	}
	if (num >= lim) break;
    }
    Dprintf("Did %d actions in all, out of %d max.\n", num, lim);
    return num;
}

unit_still_acting(unit, checkwaiting)
Unit *unit;
int checkwaiting;
{
    return (alive(unit)
	    && inside_area(unit->x, unit->y)
	    && (unit->side
		&& unit->side->ingame
		&& !unit->side->finishedturn)
	    && (unit->act
		&& unit->act->acp > 0)
	    && ((unit->plan
		&& !unit->plan->asleep
		&& !unit->plan->reserve
		&& (checkwaiting ? !unit->plan->waitingfortasks : TRUE))
		|| (unit->act && unit->act->nextaction.type != A_NONE)));
}

/* Do a single unit's actions, up to the given limit or until it runs
   out of things it wants to do. */

move_one_unit_multiple(unit, lim)
Unit *unit;
int lim;
{
    int num = 0, buzz = 0, acp1 = unit->act->acp;
    int rslt;

    while ((unit->act
	    && unit->act->initacp > 0
	    && unit->act->acp > u_acp_min(unit->type))
	   && (unit->plan && !unit->plan->asleep && !unit->plan->reserve)
	   && num < lim
	   && buzz < lim) {
	if (unit->act->nextaction.type != A_NONE) {
	    /* Execute a buffered-up action. */
	    Dprintf("%s to act;  ", unit_desig(unit));
	    Dprintf("%s", action_desig(&(unit->act->nextaction)));
	    /* Execute the action directly. */
	    rslt = execute_action(unit, &(unit->act->nextaction));
	    Dprintf(" - %s\n", hevtdefns[rslt].name);
	    unit->act->nextaction.type = A_NONE;
	    gamestatesafe = FALSE;
	    ++num;
	} else if (unit->plan) {
	    if (unit->plan->waitingfortasks
		|| unit->plan->asleep
		|| unit->plan->reserve) break;
	    execute_plan(unit, 1);
	    gamestatesafe = FALSE;
	    ++buzz;
	} else {
	    run_warning("Planless \"%s\" was asked to act", unit_desig(unit));
	    ++buzz;
	}
	if (unit->act->acp == acp1 && num > 1) {
	    /* Blast the action. */
	    unit->act->nextaction.type = A_NONE;
	    /* Blast the plan. */
	    if (unit->plan) unit->plan->type = UNFORMED_PLAN;
	    if (unit->plan && probability(5)) unit->plan->asleep = TRUE;
	    Dprintf("%s was confused, starting over\n", unit_desig(unit));
	}
    }
    return num;
}

/* This explicitly finishes out a side's activity for the turn. */

finish_turn(side)
Side *side;
{
    Side *side2;

    /* Flag the side as being done for this turn. */
    side->finishedturn = TRUE;
    /* Stop counting down our time consumption. */
    side->totaltimeused += (time(0) - side->turnstarttime);
    /* Clue everybody in. */
    for_all_sides(side2) {
	/* but only if in the game? */
	update_side_display(side2, side, TRUE);
    }
    Dprintf("%s has finished its turn.\n", side_desig(side));
}

/* Take care of details that no longer require any interaction, at least
   none that can't wait until the next turn. */

finish_movement()
{
    Unit *unit;
    Side *side;

    /* Remove dead units that have been dead more than a turn. */
    for_all_sides (side) {
    	/* (should Dprintf forfeited acp and mp here) */
	flush_side_dead(side);
	/* Clear wakeup reasons for non-movers. */
	/* need to fix this so that these stay around for a turn. */
/*	for_all_side_units(side, unit) {
	    side->laststarttime = time(0);
	} */
    }
    for_all_sides(side)  {
	update_side_display(side, side, TRUE);
    }
}

test_agreements()
{
    Agreement *ag;
    Side *side, *winner;
    Unit *unit;
    int numleft = 0, numu = 0;

    for_all_agreements(ag) {
	if (ag->active) {
	    if (symbolp(car(ag->terms))
		&& strcmp(c_string(car(ag->terms)), "last-side-wins") == 0
		&& numsides >= 2) {
    		for_all_sides(side) {
    		    numu = 0;
    		    for_all_side_units(side, unit) {
    		    	if (in_play(unit)) ++numu;
    		    }
		    if (numu == 0) {
			side_loses(side);
		    } else if (side->ingame) {
		    	++numleft;
		    	winner = side;
		    }
		}
		if (numleft == 1) {
		    side_wins(winner);
		}
	    }
	}
    }
}

/* Compute the position of the sun for this turn. */

#define lighting(x,y,snx,sny)  \
  ((distance(wrapx(x), y, snx, sny) < world.circumference / 2) ? 2 : 0)

run_sun()
{
    int daylen = world.daylength, lastsunx, lastsuny, x, y;
    Side *side;

    switch (daylen) {
      case 0:
	/* Sun is at a fixed position. */
	/* (should be possible to set explicitly somehow?) */
	sunx = area.width / 2;  suny = area.height / 2;
	daynight = TRUE;
	break;
      case 1:
	/* No sun effects at all, every place uniformly lit. */
	break;
      default:
	/* If world has a appropriate circumference, the sun moves over
	   it at a regular pace. */
	if (world.circumference >= area.width) {
	    lastsunx = sunx;  lastsuny = suny;
	    sunx = (((curturn - g_initial_day_part() + daylen) % daylen) * world.circumference) / daylen;
	    /* (should adjust for nonzero latitudes) */
	    suny = area.height / 2;
	    daynight = TRUE;
	    	/* Update the appearance of any cells whose lighting has changed. */
	    	for_all_hexes(x, y) {
	    	    if (lighting(x, y, sunx, suny) != lighting(x, y, lastsunx, lastsuny)) {
	    	    	for_all_sides(side) {
	    	    	    update_cell_display(side, x, y, FALSE);
	    	    	}
	    	    }
	    	}
	}
	break;
    }
    if (daynight) {
	Dprintf("Sun is now at %d,%d\n", sunx, suny);
    }
}

/* Seasonal change moves the average temperate up and down gradually,
   modifying it according to the terrain underneath. */

/* This should check that equator and northpole are not equal. */

run_environment()
{
    int yrlen = world.yearlength;
    int season, x, y, dir, lattemp, hextemp;
    float spoletemp, npoletemp, equatortemp;
    float poledelta, equatordelta, ntempdelta, stempdelta;

    if (yrlen > 1 /* and axial tilt... */) {
#if 0
	/* Compute the difference between summer and winter temps. */
	/* (Should be precomputed.) */
	poledelta = (g_summer_pole() - g_winter_pole()) / (yrlen / 2);
	equatordelta = (g_summer_equator() - g_winter_equator()) / (yrlen / 2);
	/* Compute where we are in the cycle of seasons. */
	season = (curturn + (yrlen - g_first_midwinter())) % yrlen;
	/* Compute temps at poles and equator. */
	if (season <= yrlen / 2) {
	    npoletemp = g_winter_pole() + season * poledelta;
	    equatortemp = g_winter_equator() + season * equatordelta;
	    spoletemp = g_winter_pole() + (yrlen/2 - season) * poledelta;
	} else {
	    npoletemp = g_winter_pole() + (yrlen - season) * poledelta;
	    equatortemp = g_winter_equator() + (yrlen - season) * equatordelta;
	    spoletemp = g_winter_pole() + (season - yrlen/2) * poledelta;
	}
	Dprintf("Temps N %f Eq %f S %f\n", npoletemp, equatortemp, spoletemp);
	/* Compute the temperature variation from one lat to the next. */
	ntempdelta = (equatortemp-npoletemp)/(world.northpole-world.equator);
	stempdelta = (equatortemp-spoletemp)/(world.equator-world.southpole);
	/* Compute the average temperature at each point in the world. */
	for (y = world.height-1; y >= 0; --y) {
	    /* For each latitude, compute the average temp. */
	    if (y >= world.equator) {
		lattemp = npoletemp
		  + ntempdelta * ((world.northpole - world.equator)
				 - (y - world.equator));
	    } else {
		lattemp = spoletemp
		  + stempdelta * ((world.equator - world.southpole)
				 - (world.equator - y));
	    }
	    for (x = 0; x < world.width; ++x) {
		if (!inside_world(x, y)) continue;
		/* Save the prev temp for a moment. */
		set_tmp1_at(x, y, temperature_at(x, y));
		if (t_tempmod(terrain_at(x, y)) > 0) {
		    /* should use tempmod value in formula. */
		    hextemp = t_average_temp(terrain_at(x, y));
		} else {
		    hextemp = lattemp;
		}
		/* Add in a random variation if specified. */
		if (g_temp_randomness() > 0) {
		    hextemp += (random(g_temp_randomness())
				- g_temp_randomness()/2);
		}
		if (elevations_defined()) {
		    /* Our basic calculations assume sea level - higher
		       elevations are colder, at a fixed rate. */
		    if (elev_at(x, y) > 0 && g_high_alt_delta() != 0) {
			hextemp -=
			  ((((long) elev_at(x, y)) * 10) / g_high_alt_delta());
		    }
		}
		/* Record the temperature of the hex. */
		set_temperature_at(x, y, hextemp);
		/* See if the display should change and report if so */
		/* (Should be controlled by other stuff.) */
#if 0
		if ((hextemp < 0 && tmp1_at(x, y) >= 0) ||
		    (hextemp >= 0 && tmp1_at(x, y) < 0)) {
		    Side *side;
		    
		    for_all_sides(side) {
			if (side_view(side, x, y) != UNSEEN) {
			    update_cell_display(side, x, y, TRUE);
			}
		    }
		}
#endif
	    }
	}
	/* (This should come *before* hex drawing!!) */
	/* Sometimes the scale of the model is such that neighboring hexes
	   influence each other. */
	if (g_moderation_range() > 0) {
	    /* only doing a range of 1... */
	    for_all_interior_hexes(x, y) {
		set_tmp1_at(x, y, temperature_at(x, y));
		for_all_directions(dir) {
		    set_tmp1_at(x, y,
				   (tmp1_at(x, y)
				    + temperature_at(x+dirx[dir], y+diry[dir])));
		}
	    }
	    for_all_interior_hexes(x, y) {
		set_temperature_at(x, y, (tmp1_at(x, y) / 7));
	    }
	}
#endif
    }
}

spying_possible()
{
    int u;

    for_all_unit_types(u) {
	if (u_spy_chance(u) > 0) return TRUE;
    }
    return FALSE;
}

/* Certain kinds of units can do spying for the side they're on. */

run_spies()
{
    Side *side;
    Unit *unit;

    if (spying_possible()) {
	Dprintf("Running spies\n");
	for_all_sides(side) {
	    for_all_side_units(side, unit) {
		if (u_spy_chance(unit->type) > 0 && in_play(unit)) {
		    if (random(10000) < u_spy_chance(unit->type)) {
			/* Spying is successful, decide what was seen. */
		    }
		}
	    }
	}
    }
}

init_random_events()
{
    /* (should fill in run/no-run slots of method table) */
    if (randomeventlist != lispnil) {
	anyrandomevents = TRUE;
#if 0
	/* this should done only during an init process of some sort */
	{
	    int u, t;
	    
	    for_all_unit_types(u) {
		if (u_revolt(u) > 0) anyrevolt = TRUE;
	    if (u_surrender(u) > 0) anysurrender = TRUE;
	    for_all_terrain_types(t) {
		if (ut_accident(u, t) > 0) anyaccident = TRUE;
		if (ut_attrition(u, t) > 0) anyattrition = TRUE;
	    }
	}
#endif
    }
}

/* Run any requested random event methods. */

run_random_events()
{
    int i;

    if (anyrandomevents) {
	Dprintf("Running random events\n");
	for (i = 0; randomeventmethods[i].name != NULL; ++i) {
	    if (randomeventmethods[i].run) {
		(*(randomeventmethods[i].fn))();
	    }
	}
    }
#if 0
    if (anyrevolt) {
	for_all_units(loop_side, unit) if (alive(unit)) unit_revolt(unit);
    }
    if (anysurrender) {
	for_all_units(loop_side, unit) if (alive(unit)) unit_surrender(unit);
    }
    if (anyaccident) {
	for_all_units(loop_side, unit) if (alive(unit)) unit_accident(unit);
    }
    if (anyattrition) {
	for_all_units(loop_side, unit) if (alive(unit)) unit_attrition(unit);
    }
    }
#endif
}

run_accidents()
{
    int t;
    Unit *unit;
    Side *loopside;

    if (0 /* anyaccident */) {
	for_all_units(loopside, unit) {
	    if (in_play(unit)) {
		t = terrain_at(unit->x, unit->y);
    		if (random(10000) < ut_accident_vanish(unit->type, t)
    		    && unit->transport == NULL) {
		    kill_unit(unit, DISASTER);
		    /* need to make a hevt */
    		}
	    }
	}
    }
}

run_revolts()
{
	Unit *unit;
	Side *loopside;

	for_all_units(loopside, unit) {
		if (in_play(unit)) {
			unit_revolt(unit);
		}
	}
}

/* A unit revolts by going back to its "true side" (which is usually set at */
/* unit creation time).  The base revolt chance is worst case if morale is */
/* variable; a unit at maximum (nonzero) morale will never revolt. */

unit_revolt(unit)
Unit *unit;
{
#if 0
    int u = unit->type, ux = unit->x, uy = unit->y, chance;
    int maxmor;
    Side *oldside = unit->side;

    if (u_revolt(u) > 0) {
	chance = u_revolt(u);
	maxmor = u_maxmorale(u);
	if (maxmor > 0) chance = (chance * (maxmor - unit->morale)) / maxmor;
	if (unit->trueside != oldside && random(10000) < chance) {
	    notify(oldside, "%s revolts!", unit_handle(oldside, unit));
	    unit_changes_side(unit, unit->trueside, CAPTURE, DISASTER);
	    see_exact(oldside, ux, uy);
	    update_cell_display(oldside, ux, uy, TRUE);
	    all_see_hex(ux, uy);
	}
    }
#endif
}

run_surrenders()
{
	Unit *unit;
	Side *loopside;

	for_all_units(loopside, unit) {
		if (in_play(unit)) {
			unit_surrender(unit);
		}
	}
}

/* Units may also surrender to adjacent enemy units, but only to a type */
/* that is both visible and capable of capturing the unit surrendering. */
/* Neutrals have to be treated differently, since they don't have a view */
/* to work from.  We sort of compute the view "on the fly". */

unit_surrender(unit)
Unit *unit;
{
#if 0
    int surrounded = TRUE;
    int u = unit->type, chance, d, x, y;
    viewdata view;
    Unit *unit2, *eunit = NULL;
    Side *us = unit->side, *es;

    if (u_surrender(u) > 0 || u_siege(u) > 0) {
	chance = 0;
	for_all_directions(d) {
	    x = wrap(unit->x + dirx[d]);  y = limit(unit->y + diry[d]);
	    if (indep(unit)) {
		if (((unit2 = unit_at(x, y)) != NULL) &&
		    (0 /* probability(u_visibility(unit2->type)) */) &&
		    (could_capture(unit2->type, u))) {
		    chance += u_surrender(u);
		    eunit = unit2;
		} else {
		    surrounded = FALSE;
		}
	    } else {
		view = side_view(us, x, y);
		if (view == EMPTY || view == UNSEEN) {
		    surrounded = FALSE;
		} else {
		    es = side_n(vside(view));
		    if (enemy_side(us, es) && could_capture(vtype(view), u)) {
			chance += u_surrender(u);
			if (unit_at(x, y)) eunit = unit_at(x, y);
		    }
		}
	    }
	}
	if (surrounded && eunit) chance += u_siege(u);
	if (random(10000) < chance) {
	    if (eunit) {
		notify(eunit->side, "%s has surrendered to you!",
		       unit_handle(eunit->side, unit));
		notify(us, "%s has surrendered to the %s!",
		       unit_handle(us, unit),
		       side_desig(eunit->side) /*plural_form(eunit->side->name)*/);
		capture_unit(eunit, unit);
	    }
	}
    }
#endif
}

run_attrition()
{
	Unit *unit;
	Side *loopside;

	for_all_units(loopside, unit) {
		if (in_play(unit)) {
			unit_attrition(unit);
		}
	}
}

/* Attrition only takes out a few hp at a time, but can be deadly... */
/* Occupants lost only if unit lost (do we care?) */

unit_attrition(unit)
Unit *unit;
{
    int u = unit->type, hipart, lopart, dmg;
    Side *us = unit->side;

    hipart = ut_attrition(u, terrain_at(unit->x, unit->y)) / 100;
    lopart = ut_attrition(u, terrain_at(unit->x, unit->y)) % 100;
    dmg = hipart + probability(lopart);
    if (dmg > 0) {
	if (unit->hp <= dmg) {
	    kill_unit(unit, DISASTER);
	} else {
	    /* (should be a version of "hit_unit"?) */
	    unit->hp -= dmg;
	}
    }
}

int tmpexcess;

/* We've "found" what we were searching when the excess to distribute is gone. */

excess_left(x, y)
int x, y;
{
	return (tmpexcess > 0);
}

try_transfer_to_cell(x, y)
int x, y;
{
	Unit *unit2, *occ;

	if (tmpexcess <= 0) return;
	for_all_stack(x, y, unit2) {
		if (in_play(unit2) && unit2->side == tmpunit->side) {
			try_transfer(tmpunit, unit2, tmpmtype);
		}
	}
	for_all_stack(x, y, unit2) {
		if (in_play(unit2) && unit2->side == tmpunit->side) {
			for_all_occupants(unit2, occ) {
				if (in_play(occ) && occ->side == tmpunit->side) {
					try_transfer(tmpunit, occ, tmpmtype);
				}
			}
		}
	}
}

/* The main routine does production, distribution, and discarding in order. */

run_economy()
{
    int u, m, r, t, amt, dist, x, y, x1, y1, x2, y2;
    int totals[MAXMTYPES];
    Unit *unit, *ounit;
    Side *loopside;

    if (nummtypes == 0) return;
    /* (should find other reasons not to run this) */
    Dprintf("Running economy\n");
    for_all_material_types(m) totals[m] = 0;
    /* Make new materials but don't clip to storage capacity yet. */
    for_all_units(loopside, unit) {
    	if (in_play(unit) && completed(unit)) {
	    u = unit->type;
	    for_all_material_types(m) {
		t = terrain_at(unit->x, unit->y);
		amt = (um_produce(u, m) * ut_productivity(u, t)) / 100 +
		  (probability((um_produce(u, m) * 
				ut_productivity(u, t)) % 100) ? 1 : 0);
		unit->supply[m] += amt;
		if (Debug && unit->supply[m] > um_storage(u, m))
		  totals[m] += (unit->supply[m] - um_storage(u, m));
	    }
	}
    }
    Dprintf("Overflow is");
    for_all_material_types(m) Dprintf(" %d", totals[m]);
    Dprintf("\n");
    /* Move stuff around - try to get rid of any excess. */
    for_all_units(loopside, unit) {
	if (in_play(unit) && !indep(unit)) {
	    u = unit->type;
	    for_all_material_types(m) {
		if (unit->supply[m] > um_storage(u, m)
		    && (dist = um_outlength(u, m)) >= 0) {
		    	tmpunit = unit;
		    	tmpmtype = m;
		    	tmpexcess = unit->supply[m] - um_storage(u, m);
			search_and_apply(unit->x, unit->y, dist, excess_left,
					 &x1, &y1, 1, try_transfer_to_cell, 999999);
		}
	    }
	}
    }
    /* Throw away excess that can't be stored anywhere. */
    for_all_material_types(m) totals[m] = 0;
    for_all_units(loopside, unit) {
	u = unit->type;
	for_all_material_types(m) {
		if (Debug && unit->supply[m] > um_storage(u, m))
		  totals[m] += (unit->supply[m] - um_storage(u, m));
	    unit->supply[m] = min(unit->supply[m], um_storage(u, m));
	}
    }
    Dprintf("Discarded ");
    for_all_material_types(m) Dprintf(" %d", totals[m]);
    Dprintf("\n");
}

/* Give away supplies, but save enough to stay alive for a couple turns. */

try_transfer(from, to, r)
Unit *from, *to;
int r;
{
    int save_amount;  /* Hang on to this much at least. */
    int u = from->type;
    int oldsupply = from->supply[r];
  
    save_amount = max(3 * um_consume(u, r),
		      2 * u_speed(u) * um_tomove(u, r));
    save_amount = min(save_amount, from->supply[r]);
    from->supply[r] -= save_amount;
    try_transfer_aux(from, to, r);
    from->supply[r] += save_amount;
    tmpexcess -= (oldsupply - from->supply[r]);
}

/* Material redistribution uses this routine to move supplies around
   between units far apart or on the same hex. Try to do reasonable
   things with the materials.  Net producers are much more willing to
   give away supplies than net consumers. */

try_transfer_aux(from, to, r)
Unit *from, *to;
int r;
{
    int nd, u = from->type, u2 = to->type, fromrate, torate;
    Unit *occ;

    if (from != to &&
	um_inlength(u2, r) >= distance(from->x, from->y, to->x, to->y)) {
	/* Try for the transfer only if we're below capacity. */
	if ((nd = um_storage(u2, r) - to->supply[r]) > 0) {
	    if ((um_produce(u, r) - um_consume(u, r) > 0)
		|| (survival_time(to) < 3)
		|| (um_storage(u, r) * 4 >= um_storage(u2, r))) {
		if (can_satisfy_need(from, r, nd)) {
		    transfer_supply(from, to, r, nd);
		} else if (can_satisfy_need(from, r, max(1, nd/2))) {
		    transfer_supply(from, to, r, max(1, nd/2));
		} else if (from->supply[r] > um_storage(u, r)) {
		    transfer_supply(from, to, r,
				    (from->supply[r] - um_storage(u, r)));
		}
	    } else {
		fromrate = u_speed(u) * um_tomove(u, r) * 3;
		fromrate = max(1, fromrate);
		torate = u_speed(u2) * um_tomove(u2, r) * 3;
		torate = max(1, torate);
		if ((from->supply[r] / fromrate) > (to->supply[r] / torate)) {
		    transfer_supply(from, to, r,
				    min(nd, (8 + from->supply[r]) / 9));
		} 
	    }
	}
    }
}

/* This estimates what can be spared.  Note that total transfer of requested */
/* amount is not a good idea, since the supplies might be essential to the */
/* unit that has them first.  If we're more than half full, or the request */
/* is less than 1/5 of our supply, then we can spare it. */

/* (should replace with doctrine/plan controls, check underlying terrain) */

can_satisfy_need(unit, r, need)
Unit *unit;
int r, need;
{
    return ((((2 * unit->supply[r]) > (um_storage(unit->type, r))) &&
	     (((unit->supply[r] * 4) / 5) > need)) ||
	    (need < (um_storage(unit->type, r) / 5)));
}

/* Handle base consumption by units. */

run_base_consumption()
{
    Unit *unit;
    Side *side;

    if (nummtypes == 0) return;
    if (0 /* any nonzero consumptions? */) return;
    Dprintf("Running base consumption\n");
    for_all_units(side, unit) {
      if (alive(unit) && completed(unit)) {
        unit_consumes(unit);
      }
    }
}

/* Consume the constant overhead part of supply consumption. */
/* Usage by movement is subtracted from overhead first. */

unit_consumes(unit)
Unit *unit;
{
    int u = unit->type, m, usedup;
    
    for_all_material_types(m) {
	if ((um_consume(u, m) > 0) &&
	    (unit->transport == NULL /* || u_consume_as_occupant(u) */)) {
	    /* Calculate what was already consumed by movement. */
	    usedup = 0;
	    if (unit->act != NULL) {
	    	usedup = unit->act->actualmoves * um_consume_per_move(u, m);
	    }
	    if (usedup < um_consume(u, m)) {
		unit->supply[m] -= (um_consume(u, m) - usedup);
	    }
	    if (unit->supply[m] <= 0 && !in_supply(unit, m)) {
		exhaust_supply(unit);
		return;
	    }
	}
    }
}

/* What happens to a unit that runs out of a supply.  If it can survive
   on no supplies, then there may be a few turns of grace, depending on
   how the dice roll... */

/* should use the per-material table */

exhaust_supply(unit)
Unit *unit;
{
#if 0
    if (probability(100 - u_survival(unit->type))) {
	/* occupants should try to escape? */
	kill_unit(unit, STARVATION);
    }
#endif
}

/* Check if the unit has ready access to a source of supplies. */

/* (should be more sophisticated and account for supply lines etc) */

in_supply(unit, m)
Unit *unit;
int m;
{
    if (unit->transport != NULL) {
    	if (unit->transport->supply[m] > 0) return TRUE;
    }
    return FALSE;
}

/* Some types of units can become completed and grow to full size
   automatically after awhile. */

run_self_builds()
{
    Unit *unit;
    Side *loopside;

    if (0 /* this step ever does anything? */) return;
    Dprintf("Running self builds\n");
    for_all_units(loopside, unit) {
	if (!fullsized(unit)
	    && u_cp_per_self_build(unit->type) > 0
	    && unit->cp >= u_cp_to_self_build(unit->type)) {
	    unit->cp += u_cp_per_self_build(unit->type);
	    if (unit->cp > u_cp(unit->type)) unit->cp = u_cp(unit->type);
	    if (completed(unit)) {
		make_unit_complete(unit);
	    } else {
	    	/* Let the player at least know that progress was made. */
		update_unit_display(unit->side, unit, TRUE);
	    }
	}
    }
}

/* See if it's time for any scheduled arrivals to appear. */

run_reinforcements()
{
    int nx, ny;
    Unit *unit, *transport;
    Side *loopside;

    if (0 /* this step never does anything? */) return;
    Dprintf("Running reinforcements\n");
    for_all_units(loopside, unit) {
    	/* See if is now time for a unit to appear. */
    	if (unit->cp < 0 && (- unit->cp) <= curturn) {
    		unit->cp = u_cp(unit->type);  /* forces to be full size! */
    		nx = (- unit->prevx);  ny = (- unit->prevy);
    		/* Do the usual steps to place the unit. */
		if (inside_world(nx, ny)) {
			if (can_occupy_cell(unit, nx, ny)) {
			    enter_hex(unit, nx, ny);
			} else {
			    /* Search this cell for units to enter. */
			    for_all_stack(nx, ny, transport) {
				if (unit->side == transport->side
				    && can_occupy(unit, transport)) {
				    enter_transport(unit, transport);
				    break;
				}
			    }
			    /* We got a problem, make the unit wait for next turn. */
			    unit->cp = (- curturn);
			}
		} else {
			/* loc of reinforcement is messed up */
		}
    	}
    }
}

/* Some types of units recover lost hp spontaneously. */

run_hp_recovery()
{
    int u, hprecovery, oldhp;
    Unit *unit;
    Side *loopside;

    if (0 /* this step ever does anything? */) return;
    Dprintf("Running hp recovery\n");
    for_all_units(loopside, unit) {
	u = unit->type;
	if (unit->hp < u_hp(u) && u_hp_recovery(u) > 0) {
	    oldhp = unit->hp;
	    hprecovery = u_hp_recovery(u);
	    unit->hp += hprecovery / 100;
	    /* Maybe recover an additional hit point. */
	    if ((hprecovery % 100) != 0) {
		if (probability(hprecovery % 100)) ++unit->hp;
	    }
	    if (unit->hp > u_hp(u)) unit->hp = u_hp(u);
	    /* Inform the player if the unit's hp changed. */
	    if (unit->hp != oldhp) {
		update_unit_display(unit->side, unit, TRUE);
	    }
	}
    }
}

/* Unconditional surrender. */

resign_game(side, side2)
Side *side, *side2;
{
    /* Tell everybody about the resignation. */
#if 0
    /* (get from nlang.c?) */
    notify_all(
#ifdef RUDE
	       "Those %s %s have given up!",
	       (flip_coin() ? "cowardly" : "wimpy"),
#else
	       "The %s have given up!",
#endif /* RUDE */
	       side->pluralnoun);
    if (side2 != NULL) {
	notify_all("... and they're giving all their stuff to the %s!",
		   side2->pluralnoun);
    }
#endif
    /* Now do the usual effects of losing. */
    /* (should we pass "resigned" as the reason?) */
    side_loses(side, side2);
}

realtime_game()
{
    return (g_rt_for_game_max() > 0 || g_rt_per_side_max());
}

all_others_willing_to_quit(side)
Side *side;
{
    Side *side2;

    for_all_sides(side2) {
	if (side != side2 && !side2->willingtodraw) return FALSE; 
    }
    return TRUE;
}
