/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Definitions about units. */

/* This structure should be small, because there may be many of them.
   Unit semantics go in this structure, while unit brains go into the
   act/plan.  Test: a unit that is like a rock and can't do anything at all
   just needs basic slots, plan needn't be allocated.  Another test:
   unit should still function correctly after its current plan has been
   destroyed and replaced with another. */

typedef struct a_unit {
    short type;                /* type (army, ship, etc) */
    short id;                  /* truly unique id number */
    char *name;                /* the name, if given */
    short number;              /* semi-unique number */
    short x, y, z;             /* position of unit in world */
    struct a_side *side;       /* whose side this unit is on */
    short hp;                  /* how much more damage each part can take */
    short cp;                  /* the state of construction of this unit */
    short cxp;                 /* the combat experience of this unit */
    struct a_unit *transport;  /* pointer to transporting unit if any */
#if (MAXMTYPES > 2)
    short *supply;             /* how much supply we're carrying (dynalloc) */
#else
    short supply[MAXMTYPES];   /* how much supply we're carrying */
#endif
    short *tooling;            /* level of preparation for construction */
    struct a_actorstate *act;  /* the unit's current actor state */
    struct a_plan *plan;       /* the unit's current plan */
    Obj *hook;                 /* placeholder for optional stuff */
    /* Following slots are never saved. */
    struct a_unit *occupant;   /* pointer to first unit being carried */
    struct a_unit *nexthere;   /* pointer to fellow occupant */
    struct a_unit *prev;       /* previous unit in list of side's units */
    struct a_unit *next;       /* next unit in list of side's units */
    struct a_unit *nextactor;  /* next unit in action circle. */
    struct a_unit *prevactor;
    struct a_unit *nextgroup;
    /* (flush eventually) */
    short prevx, prevy;        /* where were we last */
} Unit;

/* Some convenient macros. */

/* Iteration over all units. */
/* Careful with this one, can run afoul of precedence rules since no
   brace to enclose the outer loop. */

#define for_all_units(s,v)  \
  for ((s) = sidelist; (s) != NULL; (s) = (s)->next) \
    for (v = (s)->unithead->next; v != (s)->unithead ; v = v->next)

/* Iteration over all units on a given side. */

#define for_all_side_units(s,v) \
    for (v = (s)->unithead->next; v != (s)->unithead ; v = v->next)

/* Iteration over all occupants of a unit (but not sub-occupants). */

#define for_all_occupants(u1,u2) \
  for ((u2) = (u1)->occupant; (u2) != NULL; (u2) = (u2)->nexthere)

#define is_unit(unit) ((unit) != NULL && is_unit_type((unit)->type))

#define alive(unit) ((unit)->hp > 0)

#define indep(unit) ((unit)->side == NULL || (unit)->side == indepside)

#define completed(unit) \
  ((unit)->cp >= (u_cp((unit)->type) / u_parts((unit)->type)))

#define fullsized(unit) \
  ((unit)->cp >= u_cp((unit)->type))

/* This is true if the unit is on the board somewhere. */

#define is_present(unit) in_play(unit)

#define in_play(unit) \
  (is_unit(unit) && alive(unit) && inside_area((unit)->x, (unit)->y))

#define is_acting(unit) is_active(unit)

#define is_active(unit) (in_play(unit) && completed(unit))

/* Types of primitive unit actions. */

typedef enum actiontype {

#undef  DEF_ACTION
#define DEF_ACTION(CODE,name,args,fn,doc) CODE,

#include "action.def"

    NUMACTIONTYPES

} ActionType;

typedef struct a_actiondefn {
    ActionType typecode;
    char *name;
    char *argtypes;
    int (*dofn)();
} ActionDefn;

typedef struct a_action {
    ActionType type;           /* the type of the action */
    long args[4];
    struct a_action *next;     /* chain to next action */
} Action;

typedef struct a_actorstate {
    short initacp;         /* how much we can still do */
    short initmp;         /* how much we can still do */
    short acp;         /* how much we can still do */
    short mp;           /* how many moves left in this turn */
    short actualactions;         /* hexes actually covered this turn */
    short actualmoves;         /* hexes actually covered this turn */
    Action nextaction;
} ActorState;

#define valid(x) ((x) == A_ANY_OK)

/* All the definitions that govern planning. */

/* A task is a single executable element of a unit's plan.  Each task type
   is something that has been found useful or convenient to encapsulate as
   a step in a plan. */

typedef enum a_tasktype {

#undef  DEF_TASK
#define DEF_TASK(CODE,name,argtypes,fn) CODE,

#include "task.def"

    NUMTASKTYPES
} TaskType;

typedef enum a_taskoutcome {
  TASK_UNKNOWN,
  TASK_FAILED,
  TASK_INCOMPLETE,
  TASK_PREPPED_ACTION,
  TASK_COMPLETE
} TaskOutcome;

typedef struct a_task {
    TaskType type;             /* the kind of activity we want to do */
    long args[4];              /* parms of the activity itself */
    short execnum;             /* how many times this has been done */
    short retrynum;            /* number of immed failures so far */
    struct a_task *next;       /* the next task to undertake */
} Task;

/* A plan is what a single unit uses to make decisions, both for itself and
   for any other units it commands.  Any unit that can act at all has a
   plan object.  A plan collects lots of unit behavior, but its most
   important structure is the task queue, which contains a list of what
   to do next, in order. */

/* Plan types distinguish several kinds of usages. */

typedef enum plantype {

#undef  DEF_PLAN
#define DEF_PLAN(CODE,name) CODE,

#include "plan.def"

    NUMPLANTYPES
} PlanType;

typedef struct a_plan {
    PlanType type;             /* general type of plan that we've got here */
    int subtype;               /* orders being followed */
    struct a_goal *maingoal;   /* the main goal of this plan */
    int active;                /* true if this plan is still being used */
    int creationtime;          /* time at which this plan was created */
    int starttime;             /* time at which this plan is to be executed */
    int endtime;               /* time to deactivate this plan */
    int asleep;                /* true if the unit is doing nothing */
    int reserve;               /* true if unit waiting until next turn */
    int waitingfortasks;
    struct a_task *tasks;      /* pointer to chain of sequential tasks */
    int priority;
    int area;
    struct a_theater *theater;
    struct a_unit *boss;       /* who gives us commands */
    struct a_unit *deputy;     /* who gets the plan if we die */
    struct a_unit *subords;
    int autotask;
    int threatalarm;
    Side *threatside;
    int threatutype;
    int supplyalarm;
    int supplytype;
    /* (flush eventually) */
    struct a_goalstate *maingoalstate;
    int x, y;
} Plan;

/* Global unit variables. */

extern Unit *unitlist;
extern Unit *tmpunit;

extern int numunits;

/* Declarations of unit-related functions. */

void init_units();
int type_can_occupy();
int can_occupy();
int can_carry();
int enter_hex();
void leave_hex();
void enter_transport();
void leave_transport();
int embark_unit();
void flush_dead_units();
void sort_units();
int low_supplies();
char *summarize_units();
Unit *create_unit();
Unit *find_unit();
int find_unit_char();
int find_unit_name();
char *random_unit_name();
char *unit_handle();
char *short_unit_handle();
char *utype_name_n();
char *unit_desig();
Unit *first_unit();
Unit *next_unit();
void insert_unit();
void delete_unit();
int evaluate_hex();
int maximize_worth();

Unit *find_next_unit();
Unit *find_prev_unit();
Unit *find_next_actor();
Unit *find_prev_actor();

char *actorstate_desig();

/* (following should be elsewhere?) */

void free_plan();
char *plan_desig();

void init_tasks();
Task *create_task();
void free_task();
char *task_desig();

extern ActionDefn actiondefns[];

Action *create_action();
char *action_desig();

/* Declare all the action functions. */

#undef  DEF_ACTION
#define DEF_ACTION(code,name,args,FN,doc) int FN();

#include "action.def"
